#include <assert.h>

#include <Xm/ScrolledW.h>
#include <Xm/Form.h>
#include <Xm/Label.h>

#include <libtools.h>

#include "signal.h"


static bool done = TRUE;


static void SignalHandler (Widget widget, SIGNAL *Data, XEvent *Event)
{
  switch (Event->type)
  {
    case MotionNotify :
      {
	XMotionEvent *Motion = (XMotionEvent *)Event;

	if (! done)
	{
	  Data->Select (Data->Sel1, Data->Sel2);  // clean actual selection
	  Data->Select (Data->Sel1, Motion->x);   // draw new selection
	}
      }
      break;

    case ButtonPress :
      {
	XButtonEvent *Button = (XButtonEvent *)Event;

	if (Data->Sel1 >= -1)  // Clear old selection, if exists
	  Data->Select (Data->Sel1, Data->Sel2);
	Data->Select (Button->x, Button->x); // Init new selection
	done = FALSE;
      }
      break;

    case ButtonRelease :
      done = TRUE;
      break;

    case Expose :
      {
	XExposeEvent *expose = (XExposeEvent *)Event;

	if (Data->Pic != None)
	  XCopyArea (XDisplay, Data->Pic, expose->window, copyGC,
		     expose->x, expose->y,  expose->width, expose->height,
		     expose->x, expose->y);
      }
      break;

  }
}


SIGNAL::SIGNAL (const char *Name, int Size,
		Widget Parent, Widget Top, TYPE Type /* = CONTINOUS */) :
  Pic    (None),
  Sel1   (-1),
  Sel2   (-1),
  Size   (Size),
  SizeX  (Size * 3),
  SizeY  (25),
  ZoomX  (6),
  ZoomY  (20.0),
  Type   (Type),
  bg     (TOOLGetColor("Grey50")),
  fg     (TOOLGetColor("Yellow")),
  Index  (0),
  MidY   (int(0.5 * SizeY)),
  Label  (NULL)
{
  XmString LabelStr;

  SetName (Name);

  Values = new double [Size];

  Form = XtVaCreateManagedWidget ("form", xmFormWidgetClass, Parent,
				  XmNtopAttachment, XmATTACH_WIDGET,
				  XmNtopWidget,     Top,
				  XmNtopOffset,     2,
				  NULL);
  LabelStr = XmStringLtoRCreate (this->Name, NORMAL_FONT);
  Label = XtVaCreateManagedWidget ("label", xmLabelWidgetClass, Form,
				   XmNlabelString,      LabelStr,
				   XmNtopAttachment,    XmATTACH_FORM,
				   XmNleftAttachment,   XmATTACH_FORM,
				   XmNbottomAttachment, XmATTACH_FORM,
				   XmNalignment,        XmALIGNMENT_BEGINNING,
				   XmNwidth,            80,
				   NULL);
  Area = XtVaCreateManagedWidget ("area", xmDrawingAreaWidgetClass, Form,
				  XmNleftAttachment, XmATTACH_WIDGET,
				  XmNleftWidget,     Label,
				  XmNleftOffset,     5,
				  XmNwidth,          SizeX,
				  XmNheight,         SizeY,
				  NULL);
  XtAddEventHandler (Area,
		     ExposureMask | EnterWindowMask | LeaveWindowMask | 
		     ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
		     StructureNotifyMask | VisibilityChangeMask | 
		     ResizeRedirectMask,
		     TRUE, (XtEventHandler)SignalHandler, (SIGNAL *)this);
  Win = XtWindow (Area);

  Clear ();
}


SIGNAL::~SIGNAL ()
{
  if (Values != NULL)
    delete Values;
}


void SIGNAL::Clear ()
{
  memset (Values, 0, sizeof(double) * Size);

  Redraw ();
}


void SIGNAL::AddValue (double Value)
{
  assert (Index < Size);

  SetValue (Index, Value);
  Index++;
}


void SIGNAL::SetName (const char *Name)
{
  XmString LabelStr;

  strncpy (this->Name, Name, 19);
  this->Name[19] = '\0';
  if (Label != NULL)
  {
    LabelStr = XmStringLtoRCreate (this->Name, NORMAL_FONT);
    XtVaSetValues (Label, XmNlabelString, LabelStr, 
		          XmNwidth,       80,
		          NULL);
  }
}  


void SIGNAL::SetValue (int Index, double Value)
{
  int    x1, x2;
  double Last;

  assert (Index >= 0 && Index < Size);

  Values[Index] = Value;

  x1 = int (ZoomX * MAX(Index-1, 0));
  x2 = int (ZoomX * MIN(Index+1, Size));

  // clear old values
  if (x2 > x1)
  {
    XSetForeground (XDisplay, copyGC, bg);
    XFillRectangle (XDisplay, Pic, copyGC, x1+1, 0, x2-x1-1, SizeY);

    XSetForeground (XDisplay, copyGC, TOOLGetColor("White"));
    XDrawPoint (XDisplay, Pic, copyGC, int(x1+ZoomX), SizeY-1);
  }

  if (Index > 0)
    Last = Values[Index-1];
  else
    Last = Value;

  XSetForeground (XDisplay, copyGC, fg);
  if (Type == DISCRETE)
  {
    if (Index > 0)
      XDrawLine (XDisplay, Pic, copyGC, 
		 int((Index-1)*ZoomX), SizeY-2-int(Last*ZoomY),
		 int( Index   *ZoomX), SizeY-2-int(Last*ZoomY));
    if (Value != Last)
      XDrawLine (XDisplay, Pic, copyGC, 
		 int(Index*ZoomX), SizeY-2-int(Last *ZoomY), 
		 int(Index*ZoomX), SizeY-2-int(Value*ZoomY));
    XDrawLine (XDisplay, Pic, copyGC, 
	       int( Index   *ZoomX), SizeY-2-int(Value*ZoomY),
	       int((Index+1)*ZoomX), SizeY-2-int(Value*ZoomY));
  }
  else
  {
    if (Index > 0)
      XDrawLine (XDisplay, Pic, copyGC,
		 int((Index-1)*ZoomX), MidY-int(Last *ZoomY),
		 int( Index   *ZoomX), MidY-int(Value*ZoomY));
    if (Index < Size-1)
      XDrawLine (XDisplay, Pic, copyGC,
		 int(Index   *ZoomX),  MidY-int(Last           *ZoomY),
		 int((Index+1)*ZoomX), MidY-int(Values[Index+1]*ZoomY));
  }
}


void SIGNAL::SetZoomX (double FactorX)
{
  ZoomX = 3 * FactorX;
  Redraw ();
}


void SIGNAL::SetZoomY (double FactorY)
{
  ZoomY = 3 * FactorY;
  Redraw ();
}


void SIGNAL::SetColor (COLOR Color)
{
  switch (Color)
  {
    case BLACK : 
      fg = TOOLGetColor ("White");
      break;
    case GREY : 
      fg = TOOLGetColor ("Grey50");
      break;
    case WHITE :
      fg = TOOLGetColor ("White");
      break;
    case RED : 
      fg = TOOLGetColor ("Red");
      break;
    case ORANGE : 
      fg = TOOLGetColor ("Orange");
      break;
    case GREEN : 
      fg = TOOLGetColor ("Green");
      break;
    case BLUE :
      fg = TOOLGetColor ("Blue");
      break;
    case VIOLETT : 
      fg = TOOLGetColor ("Violet");
      break;
    case BROWN : 
      fg = TOOLGetColor ("Brown");
      break;
    case YELLOW : 
    default :
      fg = TOOLGetColor ("Yellow");
  }
  Redraw ();
}


void SIGNAL::SetPos (int Pos)
{
  assert (Pos >= 0 && Pos < Size);

  Index = Pos;
}


void SIGNAL::Select (int From, int To)
{
  Sel1 = From;
  Sel2 = To;

  XSetForeground (XDisplay, xorGC, TOOLGetColor("Blue") ^ bg);
  XFillRectangle (XDisplay, Pic, xorGC, MIN(From,To), 0, 
		                        abs(Sel2-Sel1), SizeY);
  Refresh ();
}


double SIGNAL::GetValue (int Index) const
{
  assert (Index >= 0 && Index < Size);

  return Values[Index];
}


double SIGNAL::GetZoomX () const
{
  return ZoomX / 3;
}


double SIGNAL::GetZoomY () const
{
  return ZoomY / 3;
}


void SIGNAL::Refresh ()
{
  // Zoom bercksichtigen
  XCopyArea (XDisplay, Pic, XtWindow(Area), copyGC, 0, 0, SizeX, SizeY, 0, 0);
}


Widget SIGNAL::GetWidget ()
{
  return Form;
}


void SIGNAL::Update ()
{
  int count = 0;

  SizeX = int (Size * ZoomX);

  XtVaSetValues (Area, XmNwidth, SizeX, NULL);

  if (Pic != None)
    XFreePixmap (XDisplay, Pic);

  Pic = XCreatePixmap (XDisplay, XtWindow(Area),
			         SizeX, SizeY, TOOLGetDepth());

  XSetForeground (XDisplay, copyGC, bg);
  XFillRectangle (XDisplay, Pic, copyGC, 0, 0, SizeX, SizeY);

  XSetForeground (XDisplay, copyGC, TOOLGetColor("White"));
  while (count < SizeX)
  {
    XDrawPoint (XDisplay, Pic, copyGC, count, SizeY-1);
    count += int(ZoomX);
  }
}


void SIGNAL::Redraw ()
{ 
  int    count;
  double Last;

  Update ();

  XSetForeground (XDisplay, copyGC, fg);

  Last = Values[0];
  if (Type == DISCRETE)
    for (count = 0; count < Size; count++)
    {  
      if (Values[count] != Last)
	XDrawLine (XDisplay, Pic, copyGC, 
		   int(count*ZoomX), SizeY-2-int(Last*ZoomY), 
		   int(count*ZoomX), SizeY-2-int(Values[count]*ZoomY));
      XDrawLine (XDisplay, Pic, copyGC, 
		 int(count    *ZoomX), SizeY-2-int(Values[count]*ZoomY),
		 int((count+1)*ZoomX), SizeY-2-int(Values[count]*ZoomY));
      Last = Values[count];
    }
  else
    for (count = 0; count < Size; count++)
    {
      XDrawLine (XDisplay, Pic, copyGC, 
		           int(count*ZoomX),     int(MidY-Last), 
		           int((count+1)*ZoomX), int(MidY-Values[count]));
      Last = Values[count];
    }
      

  Refresh ();
}
